#version 420
#extension GL_ARB_explicit_uniform_location : enable
#extension GL_ARB_separate_shader_objects : enable

in vec3 toEyeDirTS;
in vec2 texCoords;
in vec3 positionVS;
in vec3 positionOS;
in vec3 normalVS;
in vec3 tangentOS;
in vec3 bitangentOS;
in vec3 normalOS;

layout(location = 0) out vec4 outColor;
layout(location = 1) out vec3 outNormal;
layout(location = 2) out vec3 outDepth;

uniform mat4 model;
uniform mat4 view;
uniform vec3 cameraPosition;

uniform sampler2D tex;
uniform sampler2D normalMap;
uniform sampler2D bloomMap;
uniform sampler2D bumpMap;

uniform float materialShininess;
uniform vec3 materialSpecularColor;
uniform float nmStrength = 1.0;
uniform float bump_scale = 0.03;

uniform int doNormalMapping = 1;

uniform sampler1D lightPositions;
uniform sampler1D lightColors;
uniform sampler1D lightDirections;

uniform mat4 depthVP;
uniform sampler2DShadow shadowMap;
uniform bool doShadowing;

float bias = 0.001;
vec2 poissonDisk[4] = vec2[] (
  vec2( -0.94201624, -0.39906216 ),
  vec2( 0.94558609, -0.76890725 ),
  vec2( -0.094184101, -0.92938870 ),
  vec2( 0.34495938, 0.29387760 )
);

vec3 calcBlinnPhong(vec3 posToLight, vec3 posToCamera, vec3 normal, vec3 color)
{
	float n_dot_l = max(0.0, dot(normal, posToLight));
	vec3 half_vec = normalize(posToLight + posToCamera);
	float n_dot_h = max(0.0, dot(normal, half_vec));
	float spec_pwr = pow(n_dot_h, materialShininess);			// materialShininess
	
	vec3 diffuse = (1.0 - materialShininess) * color * n_dot_l;		// or diffuseReflectivity
	vec3 specular = vec3(materialShininess) * spec_pwr;		// or specularReflectivity
	return diffuse + specular;
}

vec3 calcNormalNMTS()
{
	vec3 normalNM = normalize(texture(normalMap, texCoords).rgb * 2.0 - 1.0);
	vec3 normalNMTS = mix(vec3(0.0, 0.0, 1.0), normalNM, nmStrength);
	return normalize(normalNMTS);
}

vec3 calcNormalNMVS(vec2 coords)
{
	mat3 matrixTStoOS = mat3(tangentOS, bitangentOS, normalOS);
	mat3 matrixOStoTS = transpose(matrixTStoOS);
	mat3 matrixVStoTS = matrixOStoTS * mat3(inverse(view * model));
	mat3 matrixTStoVS = transpose(matrixVStoTS);
	
	vec3 normalNM = normalize(texture(normalMap, coords).rgb * 2.0 - 1.0);
	vec3 diffTS = vec3(normalNM.x, -normalNM.y, normalNM.z) - vec3(0.0, 0.0, 1.0);
	vec3 diffVS = matrixTStoVS * diffTS;
	vec3 normalNMVS = mix(normalVS, normalVS + diffVS, nmStrength);
	
	return normalize(normalNMVS);
}

vec4 calcLighting(vec3 texColor, vec2 coords)
{
	vec3 color = vec3(0.0, 0.0, 0.0);

	vec3 normalNMTS = calcNormalNMTS();

	vec3 toCameraVS = normalize((view * vec4(cameraPosition, 1.0)).xyz - positionVS.xyz);
	vec3 normalNMVS = calcNormalNMVS(coords);

	int nrLights = textureSize(lightPositions, 0);

	vec3 posToLight;
	float dist;
	float attenuation;
	vec3 lightIntensity;
	float coneAngle;
	vec3 lightPosition;
	vec3 lightColor;
	vec3 coneDir;

	for (int i = 0; i < nrLights; i++) 
	{
		float i_float = float(i);
		float texPos = i_float/(nrLights-1);

		attenuation = texture(lightColors, texPos).a;
		lightPosition = texture(lightPositions, texPos).xyz;
		lightColor = texture(lightColors, texPos).xyz;

		// ambient light
		if (lightPosition == vec3(0.0, 0.0, 0.0))
		{
			color += texColor * lightColor;
		}

		// directional light
		else if (attenuation == 0.0)
		{
			float visibility = 1.0f;
			if (doShadowing)
			{
				visibility = 0.0f;
				vec4 shadowCoords = depthVP * model * vec4(positionOS, 1.0);
				for (int i = 0; i < 4; i++){
					visibility += texture(shadowMap, vec3(shadowCoords.xy + poissonDisk[i]/8000.0, shadowCoords.z/shadowCoords.w - bias));
				}
				visibility /= 4;
			}

			posToLight = normalize((view * vec4(lightPosition, 0.0)).xyz);

			if (doNormalMapping == 1) {
				color += visibility * lightColor * calcBlinnPhong(posToLight, toCameraVS, normalNMVS, texColor);
			} else {
				color += visibility * lightColor * calcBlinnPhong(posToLight, toCameraVS, normalVS, texColor);
			}
		}

		// cone light
		else
		{
			posToLight = normalize((view * vec4(lightPosition, 1.0)).xyz - positionVS);

			coneDir = normalize((view * vec4(texture(lightDirections, texPos).xyz, 0.0)).xyz);
			float lightToSurfaceAngle = degrees(acos(dot(-posToLight, coneDir)));

			if (texture(lightDirections, texPos).w > lightToSurfaceAngle)
			{
				dist = distance(positionVS, (view * vec4(lightPosition, 1.0)).xyz);
				attenuation = texture(lightColors, texPos).a * dist * dist;
				lightIntensity = lightColor / attenuation;
			
				vec3 result;
				if (doNormalMapping == 1) {
					result = lightIntensity * calcBlinnPhong(posToLight, toCameraVS, normalNMVS, texColor);
				} else {
					result = lightIntensity * calcBlinnPhong(posToLight, toCameraVS, normalVS, texColor);
				}
				if (result.x > 0.0)
				{
					color += result;
				}
			}
		}
	}

	return vec4(color, 0.0);
}

vec2 calcParallax()
{
	mat3 matrixTStoOS = mat3(tangentOS, bitangentOS, normalOS);
	mat3 matrixOStoTS = transpose(matrixTStoOS);
	mat3 matrixVStoTS = matrixOStoTS * mat3(inverse(view * model));

	float height = texture(bumpMap, texCoords).r;
	height = height * bump_scale - 0.5 * bump_scale;
	vec3 toCameraVS = normalize((view * vec4(cameraPosition, 1.0)).xyz - positionVS.xyz);
	vec3 toCameraTS = matrixVStoTS * toCameraVS;
	return texCoords + (height * toCameraTS.xy);
}

void main()
{
	vec2 coords = texCoords;
	if (doNormalMapping == 1)
	{
		coords = calcParallax();
	}
	outNormal = (calcNormalNMVS(coords) + 1) / 2;
	//outNormal = (normalVS + 1) / 2;
	float depthVal = gl_FragCoord.z;
	depthVal -= 0.95;
	depthVal *= 20;
	outDepth = vec3(depthVal);

	vec3 texColor = texture(tex, coords).rgb;
	outColor = calcLighting(texColor, coords);

	float bloomFactor = texture(bloomMap, coords).r;
	outColor = mix(outColor, vec4(texColor, 0.0), bloomFactor);
}